from collections import defaultdict
from time import mktime


class PathInfo:
    def __init__(self, key):
        if key and key[-1] == '/':
            key = key[:-1]
        self.key = key
        self.storage_classes = defaultdict(StorageClassInfo)

    def __repr__(self):
        return '<PathInfo: %s, %d bytes, %d files>' % (
            self.key,
            self.size,
            self.file_count,
        )

    @property
    def file_count(self):
        return sum([x.file_count for x in self.storage_classes.values()])

    @property
    def size(self):
        return sum([x.size for x in self.storage_classes.values()])

    @property
    def ts_min(self):
        return min([x.ts_min for x in self.storage_classes.values()])

    @property
    def ts_max(self):
        return max([x.ts_max for x in self.storage_classes.values()])

    def add_file(self, file):
        self.storage_classes[file['StorageClass']].add_file(file)

    def add_directory(self, directory):
        for storage_class, info in directory.storage_classes.items():
            self.storage_classes[storage_class].add(info)

    @property
    def parent(self):
        # reverse the string, split it by / once, grab the second segment
        # and reverse again
        print([self.key])
        if not self.key:
            return None
        if "/" not in self.key:
            return ""
        return self.key[::-1].split('/', 1)[1][::-1]

    def to_json(self):
        return {
            'key': self.key,
            'data': dict((x[0], x[1].to_json()) for x in self.storage_classes.items())
        }


class StorageClassInfo:
    def __init__(self):
        self.file_count = 0
        self.size = 0
        self.ts_min = None
        self.ts_max = None

    @staticmethod
    def __timestamp_from_datetime(dt):
        """Python 2.7 doesn't have a datetime.timestamp method"""
        return mktime(dt.timetuple())

    @staticmethod
    def __func_none(f, a, b):
        """Return f(a,b), unless a or b are None, in which case return the other one.
        Usefull for min/max"""
        if a is None:
            return b
        elif b is None:
            return a
        return f(a, b)

    def add(self, new):
        self.file_count += new.file_count
        self.size += new.size
        self.ts_min = StorageClassInfo.__func_none(min, self.ts_min, new.ts_min)
        self.ts_max = StorageClassInfo.__func_none(max, self.ts_max, new.ts_max)

    def add_file(self, new_file):
        self.file_count += 1
        self.size += new_file['Size']
        self.ts_min = StorageClassInfo.__func_none(
            min, self.ts_min,
            StorageClassInfo.__timestamp_from_datetime(new_file['LastModified'])
        )
        self.ts_max = StorageClassInfo.__func_none(
            max, self.ts_max,
            StorageClassInfo.__timestamp_from_datetime(new_file['LastModified'])
        )

    def __repr__(self):
        return '<StorageClassInfo: %d, %d>' % (
            self.file_count,
            self.size,
        )

    def to_json(self):
        return {
            'file_count': self.file_count,
            'size': self.size,
            'ts_min': int(self.ts_min),
            'ts_max': int(self.ts_max),
        }



from .pathinfo import PathInfo


class Node:
    def __init__(self, bucket, prefix, s3_client, progress_bar, include_files=False):
        self.bucket = bucket
        self.prefix = prefix
        self.s3_client = s3_client
        self.include_files = include_files
        self.files = []
        self.nodes = []
        self.progress_bar = progress_bar
        self.data = None

    @classmethod
    def from_dict(cls, obj):
        result = Node()
        self.prefix = obj['Key']


    @property
    def list(self):
        first = True
        result = {}
        while first or 'NextContinuationToken' in result:
            if first:
                result = self.s3_client.list_objects_v2(
                    Bucket=self.bucket,
                    Prefix=self.prefix,
                    Delimiter='/'
                )
            else:
                result = self.s3_client.list_objects_v2(
                    Bucket=self.bucket,
                    Prefix=self.prefix,
                    Delimiter='/',
                    ContinuationToken=result['NextContinuationToken']
                )
            first = False
            for prefix in result.get('CommonPrefixes', []):
                yield prefix

            for content in result.get('Contents', []):
                yield content

    def disk_usage(self):
        """ Recursively compute disk usage for all children
        """
        # one key/value for each storage class
        self.data = PathInfo(self.prefix)

        self.progress_bar.update("s3://%s/%s" % (self.bucket, self.prefix))
        for key in self.list:
            if 'Prefix' in key:
                # this is a subdirectory, going deeper
                node = Node(self.bucket, key['Prefix'], self.s3_client, self.progress_bar, self.include_files)
                last_item = None
                for item in node.disk_usage():
                    last_item = item
                    yield item
                if last_item:
                    # only add the last item in a directory to the total count
                    self.data.add_directory(last_item)
            else:
                path_info = PathInfo(key['Key'])
                path_info.add_file(key)
                self.data.add_file(key)
                if self.include_files:
                    yield path_info

        yield self.data

    def __repr__(self):
        return 's3://%s/%s' % (self.bucket, self.prefix)



from setuptools import setup

setup(
    name='cdu',
    version='0.2.0a1',
    packages=['cdu'],
    description="Cloud Storage Disk Usage Analyzer",
    author="Adrian Mester",
    author_email="mesteradrian@gmail.com",
    url="https://github.com/ilogik/cdu",
    keywords=["aws", "s3", "boto"],
    entry_points={
        'console_scripts': [
            'cdu = cdu.main:main',
            'cdun = cdun.main:main',
        ]
    },
    install_requires=[
        'boto3',
        'six',
    ]
)



import time


class Progress:
    # don't update more than once every 2 seconds
    MIN_REFRESH_INTERVAL = 2

    def __init__(self, stream):
        self.count = 0
        self.text = ""
        self.last_lengths = [0, 0]
        self.stream = stream
        self._last_updated = None
        self.stream.write('\n\n')

    def update(self, new_text):
        self.count += 1
        self.text = new_text
        self.refresh()

    def refresh(self):
        now = time.time()
        if self._last_updated is not None and now - self._last_updated <= self.MIN_REFRESH_INTERVAL:
            return
        self._last_updated = now
        self.stream.write("\033[F\033[F")
        for length in self.last_lengths:
            self.stream.write(" " * length)
            self.stream.write("\n")
        self.stream.write("\033[F\033[F")
        messages = [
            '%d done' % self.count,
            self.text
        ]
        for m in messages:
            self.stream.write("%s\n" % m)
        self.last_lengths = [len(x) for x in messages]


"""Utility to calculate disk usage on cloud services"""

import sys
import argparse
from six.moves.urllib.parse import urlparse
import boto3
import json
from .node import Node
from .progress import Progress
import time
import gzip
import sqlite3
from datetime import datetime
import os.path

FORMAT_VERSION = 1


def main():
    """Entry-point when running the module as an executable"""

    parser = argparse.ArgumentParser(
        description="Utility to calculate disk usage on cloud services"
    )
    parser.add_argument('--include-files', dest='include_files', action='store_true')
    parser.add_argument('-p', metavar='aws_profile_name', type=str,
                        help='The name of the aws profile, defined in ~/.aws/credentials',
                        default=None, required=False)
    parser.add_argument('url', type=str, help="S3 URL to process")
    parser.set_defaults(include_files=False)
    args = parser.parse_args()

    url = urlparse(args.url)

    if url.scheme == 's3':
        session = boto3.Session(profile_name=args.p)
        s3_client = session.client('s3')

        bucket = url.netloc
        # remove the first slash
        prefix = url.path[1:]
        if prefix == '/':
            # remove the last slash as well if present
            prefix = prefix[:-1]

        output_filename = '%s_%s.cdu' % (
            bucket,
            datetime.now().strftime('%Y-%m-%d_%H-%M')
        )
        if os.path.exists(output_filename):
            raise Exception("File already exists")
        
        db = sqlite3.connect(output_filename)

        cursor = db.cursor()
        cursor.execute("CREATE table metadata (key TEXT, value TEXT)")
        cursor.executemany("INSERT INTO metadata VALUES (?,?)", (
            ('version', 1),
            ('bucket', bucket),
            ('prefix', prefix),
            ('timestamp', int(time.time()))
        ))

        cursor.execute("CREATE TABLE pathinfo (parent TEXT, info TEXT)")
        cursor.execute("CREATE INDEX parent ON pathinfo(parent)")

        node = Node(bucket, prefix, s3_client, progress_bar=Progress(sys.stderr), include_files=args.include_files)
        cursor.executemany(
            "INSERT INTO pathinfo VALUES (?,?)",
            (
                (x.parent, json.dumps(x.to_json())) for x in node.disk_usage()
            )
        )

        db.commit()
        db.close()

    else:
        raise Exception('Unsupported url: "%s". At the moment only s3:// is supported' % args.url)


if __name__ == "__main__":
    main()
